#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>
#include <errno.h>
#include <math.h>
#define cfatal(cond,msg) do { if (cond) {fprintf (stderr,(msg)); exit(1);} } while (0)

const int   hemisphere_file_size = 1399680000;
const char *hemisphere_file_name_east = "data/east";
const char *hemisphere_file_name_west = "data/west";

typedef unsigned char byte;

struct line_buffer {
    byte  * buf;
    int     system_size;
    int     line_nr;
};

struct hemisphere {
    int fd;
    byte **line_ptrs;
    struct line_buffer buffer[16];
    int    cur;
} west, east;

/* AUX */

void init_hemisphere(struct hemisphere *h, const char *fn)
{
    int i;
    
    h->fd = open (fn, O_RDONLY);
    cfatal(h->fd == -1, "Error opening file\n");
    h->line_ptrs = malloc (21600 * sizeof(byte*));
    for (i=0; i<21600; i++)
	h->line_ptrs[i] = 0;
    for (i=0; i<16; i++)
    {
	h->buffer[i].buf = 0;
	h->buffer[i].system_size = 0;
	h->buffer[i].line_nr = -1;
    }
    h->cur = 0;
}

void init()
{
    init_hemisphere(&east, hemisphere_file_name_east);
    init_hemisphere(&west, hemisphere_file_name_west);
}

void load_line(struct hemisphere *h, int nr)
{
    byte *buf;
    int ofs, sz;
    int i;

    if (h->buffer[h->cur].buf)
    {
	munmap (h->buffer[h->cur].buf, h->buffer[h->cur].system_size);
	h->line_ptrs[h->buffer[h->cur].line_nr] = (byte*)0;
    }
    
    ofs = 3 * 21600 * nr;
    sz = 3 * 21600;
    i = ofs % 4096;
    ofs -= i;
    sz  += i;
    
    buf = mmap(0, sz, PROT_READ, MAP_PRIVATE, h->fd, ofs);
    cfatal (!buf, "Can't mmap() a line");

    h->buffer[h->cur].buf = buf;
    h->buffer[h->cur].system_size = sz;

    h->buffer[h->cur].line_nr = nr;
    h->line_ptrs[nr] = &h->buffer[h->cur].buf[i];

    h->cur ++;
    h->cur %= 16;
}

void print_ppm_headers(FILE *file, int size_x, int size_y)
{
    fprintf(file, "P6\n%d %d\n255\n", size_x, size_y);
    fflush(file);
}

void force_write(int fd, char *buf, int size)
{
    int i = 0;
    int res;
    
    while (i < size)
    {
	res = write (fd, buf + i, size - i);
	if (res == -1)
	    cfatal (errno != EINTR, "Error outputting data\n");
	else
	    i += res;
    }
}

inline byte *get_value(int x, int y)
{
    static byte zero [] = {0, 0, 0};
    if (y < 0 || y >= 21600)
	return zero;
/* C is truly braindead. The only correct answer for -2 % 3 is 1
 * but C somehow thinks it's -2
 */
    x = (x + 21600 * 2) % (21600 * 2);
    if (x >= 21600) {
	if (!west.line_ptrs[y])
	    load_line(&west, y);
	return &west.line_ptrs[y][3 * (x - 21600)];
    } else {
	if (!east.line_ptrs[y])
	    load_line(&east, y);
	return &east.line_ptrs[y][3 * x];
    }
}

inline void bilinear_put(byte * buf, int x, int tx, int ty, float px, float py)
{
    byte * ul = get_value(tx, ty);
    byte * ur = get_value(tx, ty + 1);
    byte * ll = get_value(tx + 1, ty);
    byte * lr = get_value(tx + 1, ty + 1);
    int i;

    for (i=0; i<3; i++)
    buf[x*3 + i] =  (byte)
		    ((py * lr[i] + (1-py) * ur[i]) * px +
		     (py * ll[i] + (1-py) * ul[i]) * (1-px));
}

inline void bilinear_put2(byte * buf, int x, float inx, float iny)
{
    int tx, ty;
    float px, py;
    
    px = inx - floor(inx);
    tx = (int)inx;
    py = iny - floor(iny);
    ty = (int)iny;
    
    bilinear_put(buf, x, tx, ty, px, py);
}

#define geo2pixel_x(geox) ((geox) * 21599 / M_PI)
#define geo2pixel_y(geoy) ((M_PI_2 - (geoy)) * 21599 / M_PI)
#define d2r(d) ((d) * (M_PI / 180.0))

/* PROJECTIONS */

void cyl_equidistant(int sizex, int sizey, float north, float west, float south, float east)
{
    byte buf[sizex * 3];
    int x, y;
    float tx, ty;
    
    print_ppm_headers(stdout, sizex, sizey);

    for (y=0; y<sizey; y++)
    {
	for (x=0; x<sizex; x++)
	{
	    tx = geo2pixel_x(west  + (east  - west ) * x / (sizex - 1));
	    ty = geo2pixel_y(north + (south - north) * y / (sizey - 1));
	    bilinear_put2(buf, x, tx, ty);
	}
	force_write(1, buf, sizex * 3);
    }
}


void cyl_equalarea(int sizex, int sizey, float north, float west, float south, float east)
{
    byte buf[sizex * 3];
    int x, y;
    float tx, ty;
    
    print_ppm_headers(stdout, sizex, sizey);

    for (y=0; y<sizey; y++)
    {
	for (x=0; x<sizex; x++)
	{
	    tx = geo2pixel_x(west  + (east  - west) * x / (sizex - 1));
	    ty = geo2pixel_y(asin(sin(north) + (sin(south) - sin(north)) * y / (sizey - 1)));
	    bilinear_put2(buf, x, tx, ty);
	}
	force_write(1, buf, sizex * 3);
    }
}

void cyl_mercator(int sizex, int sizey, float north, float west, float south, float east)
{
    byte buf[sizex * 3];
    int x, y;
    float tx, ty;
    
    print_ppm_headers(stdout, sizex, sizey);
#define forw(x) log(tan(x) + 1/cos(x))
#define back(x) (2 * atan (exp(x)) - M_PI_2)

    for (y=0; y<sizey; y++)
    {
	for (x=0; x<sizex; x++)
	{
	    tx = geo2pixel_x(west  + (east  - west) * x / (sizex - 1));
	    ty = geo2pixel_y(back(forw(north) + (forw(south) - forw(north)) * y / (sizey - 1)));
	    bilinear_put2(buf, x, tx, ty);
	}
	force_write(1, buf, sizex * 3);
    }
}

int main(int argc, char **argv)
{
    int xsize;
    int ysize;
    float north;
    float south;
    float west;
    float east;

    cfatal (argc != 8, "Usage: project <projection> <xsize> <ysize> <north> <south> <west> <east>\n");

    xsize = atoi(argv[2]);
    ysize = atoi(argv[3]);
    north = d2r(atof(argv[4]));
    south = d2r(atof(argv[5]));
    west  = d2r(atof(argv[6]));
    east  = d2r(atof(argv[7]));
    
    init ();

    if (strcmp(argv[1],"a") == 0)
	cyl_equalarea(xsize,ysize,north,west,south,east);
    else if (strcmp(argv[1],"m") == 0)
	cyl_mercator(xsize,ysize,north,west,south,east);
    else if (strcmp(argv[1],"d") == 0)
	cyl_equidistant(xsize,ysize,north,west,south,east);
    else
	cfatal(1, "Unknown projection type\n");
    return 0;
}
