#!/usr/bin/ruby

require 'RMagick'
include Math
include Magick

class Array
    def any?(&b)
	result = false
	each{|*i| result ||= b.call(*i)}
	result
    end
    def each_pair2
	return if size == 0
	lx,ly = self[-1]
	each_with_index{|(x,y),i|
	    yield(x,y,lx,ly)
	    lx,ly = x,y
	}
    end
end

PI_180 = PI / 180.0

module Projection
    class Cylindrical
	def initialize(xsize,ysize,north,south,west,east)
	    @xsize = xsize.to_f
	    @ysize = ysize.to_f
	    @north = north.to_f
	    @south = south.to_f
	    @west  = west.to_f
	    @east  = east.to_f
	    @xmul = @xsize / (@east  - @west)
	    @xadd = -@west
	end
	def point_on_map?(x,y)
	    x >= @west && x <= @east && y >= @north && y <= @south
	end
	def bb_on_map?(l,r,u,d)
	    r >= @west && l <= @east && u >= @south && d <= @north
	end
	def map_intersect(b)
	    Box.new([@west, b.l].max, [@east, b.r].min, [@north, b.u].min, [@south, b.d].max)
	end
	def cmd
	    "#{cname} #{@xsize} #{@ysize} #{@north} #{@south} #{@west} #{@east}"
	end
	def projected_box_area(box)
	    x1,y1 = project(box.l, box.u)
	    x2,y2 = project(box.r, box.d)
	    ((x1-x2) * (y1-y2)).abs
	end
    end
    class Equidistant < Cylindrical
	def initialize(xsize,ysize,north,south,west,east)
	    super(xsize,ysize,north,south,west,east)
	    @ymul = @ysize / (@south - @north)
	    @yadd = -@north
	end
	def project(long,lat)
	    [(long + @xadd) * @xmul,
	     (lat  + @yadd) * @ymul
	    ]
	end
	def cname
	    "d"
	end
    end
    class Equalarea < Cylindrical
	def initialize(xsize,ysize,north,south,west,east)
	    super(xsize,ysize,north,south,west,east)
	    @ymul = @ysize / (sin (@south * PI_180) - sin(@north * PI_180))
	    @yadd = -sin(@north * PI_180)
	end
	def project(long,lat)
	    [(long + @xadd) * @xmul,
	     (sin (lat * PI_180) + @yadd) * @ymul
	    ]
	end
	def cname
	    "a"
	end
    end
    class Mercator < Cylindrical
	def forw(lat)
	    lat = (89 + lat)/2 if lat > 89
	    lat = (-89 + lat)/2 if lat < -89
	    log(tan(lat * PI_180) + 1/cos(lat * PI_180))
	end
	def project(long,lat)
	    [(long + @xadd) * @xmul,
	     (forw (lat)    - forw(@north))/
	     (forw (@south) - forw(@north)) * @ysize
	    ]
	end
	def cname
	    "m"
	end
    end
end

class Box < Struct.new(:l,:r,:u,:d)
    def disjoint?(b)
	b.r <= l || b.l >= r || b.u >= d || b.d <= u
    end
    def fully_inside?(b)
	b.l >= l && b.r <= r && b.u <= u || b.d >= d
    end
    def overlap(b)
	[r - b.l, b.r - l].max * [d - b.u, b.d - u].max
    end
    def +(b)
	Box.new([l,b.l].min, [r,b.r].max, [u,b.u].max, [d,b.d].min)
    end
    def mx
	(l + r) / 2
    end
    def my
	(u + d) / 2
    end
end

class CityDot < Struct.new(:x,:y,:overlap_range)
    def box
	@box ||= Box.new(x-overlap_range, x+overlap_range, y-overlap_range, y+overlap_range)
	@box
    end
end

class CityLabel < Struct.new(:rx,:ry,:x,:y,:w,:h,:label,:offmax)
    def box
	l = x + rx * (w / 2 + offmax) - w/2
        r = l + w
        u = y + ry * (h / 2 + offmax) - h/2
        d = u + h
	Box.new (l,r,u,d)
    end
end

class Labels
    def initialize(image)
	@image = image
	@text = Draw.new
	@text.text_antialias = true
	@text.pointsize = 12
	@text.fill = 'white'
	@labels = []
	@citydots = []
    end
    def improve_placement(temp)
	buckets=[]
	@labels.each_with_index {|l,i|
	    b = l.box
	    conflicts = []
	    buckets.each_with_index{|(bb,labels),j|
		if not b.disjoint?(bb) && labels.any?{|k| not b.disjoint?(@labels[k].box)}
		    conflicts.push j
		end
	    }
	    lss=[i]
	    dsf=0
	    conflicts.each {|j|
	        bb,ls = buckets.delete_at (j-dsf)
		dsf += 1
		b   += bb
	        lss += ls
	    }
	    buckets.push [b, lss]
	}
	vx=[]
	vy=[]
	buckets.each{|b,ls|
	    ls.each {|i| ls.each{|j|
		next if i == j
		l = @labels[i].box
		q = @labels[j].box
		unless l.disjoint? q
		    ivx = (l.mx - q.mx)
		    ivy = (l.my - q.my)
		    ivn = sqrt(ivx * ivx + ivy * ivy)
		    o = l.overlap q
		    vx[i] ||= 0.0
		    vy[i] ||= 0.0
		    vx[i] += o * ivx / ivn
		    vy[i] += o * ivy / ivn
		end
	    }}
	}
	@labels.each_with_index{|el,i|
	    b = el.box
	    rx = el.rx
	    ry = el.ry
	    vx[i] ||= 0.0
	    vy[i] ||= 0.0
	    @citydots.each{|c|
		unless b.disjoint? c.box
		    ivx = b.mx - c.x
		    ivy = b.mx - c.y
		    ivn = sqrt(ivx * ivx + ivy * ivy)
		    vx[i] += ivx / ivn
		    vy[i] += ivy / ivn
		end
	    }
	    n = sqrt (vx[i] * vx[i] + vy[i] * vy[i])
	    if n == 0
		if rx.abs < ry.abs
		    rx = (rx > 0) ? [rx-temp, 0].max : [rx+temp, 0].min
		else
		    ry = (ry > 0) ? [ry-temp, 0].max : [ry+temp, 0].min
		end
	    else
	        rx = rx + temp * vx[i]/n
		ry = ry + temp * vy[i]/n
		rx =  1 if rx >  1
		rx = -1 if rx < -1
		ry =  1 if ry >  1
		ry = -1 if ry < -1
	    end
	    @labels[i].rx = rx
	    @labels[i].ry = ry
	}
    end
    def draw
	10.times {improve_placement 0.5}
	10.times {improve_placement 0.4}
	10.times {improve_placement 0.3}
	10.times {improve_placement 0.2}
	10.times {improve_placement 0.1}
	10.times {improve_placement 0.05}
	@labels.each {|l|
	    b = l.box
	    @text.annotate(@image, 0, 0, b.l, b.d, l.label)
	}
    end
    def add_city(size,name,long,lat)
	x,y = $proj.project (long,lat)
	if size == "b"
	    size          = 4
	    overlap_range = 6
	    offmax        = 8
	else
	    size          = 2.5
	    overlap_range = 3
	    offmax        = 5
	end

	dot = Draw.new
#	dot.antialias true
	dot.stroke 'none'
        dot.fill 'white'
	dot.circle(x, y, x, y + size)
        dot.draw @image
	
	metrics = @text.get_type_metrics(@image,name)
	@labels.push CityLabel.new(1, 0, x, y, metrics.width, metrics.ascent, name, offmax)
	@citydots.push CityDot.new(x, y, overlap_range)
    end
end

# PARSING

rivers      = false
clabels     = false
locale	    = nil
cities      = nil
map_x,map_y = 500,500
proj_n,proj_s,proj_w,proj_e = nil,nil,nil,nil
projection  = Projection::Equalarea
outfile     = nil
type_bm     = true

STDIN.each{|line|
    line.chomp!
    case line
    when /size\s*=\s*(\d+)\s+(\d+)/
	map_x = $1.to_i
	map_y = $2.to_i
    when /lat\s*=\s*(-?\d+)\s+(-?\d+)/
	proj_n = $1.to_i
	proj_s = $2.to_i
    when /long\s*=\s*(-?\d+)\s+(-?\d+)/
	proj_w = $1.to_i
	proj_e = $2.to_i
    when /background\s*=\s*(.*)/
	case $1
	when 'satellite','bluemarble'
	    type_bm = true
	when 'political','polygons'
	    type_bm = false
	else
    	    STDERR.print "Unknown map type: #{background}\n"
	    exit 1
	end
    when /locale\s*=\s*(.*)/
	locale = $1
    when "rivers"
	rivers = true
    when "clabels"
	clabels = true
    when /cities\s*=\s*(.*)/
	cities = $1
    when /out\s*=\s*(.*)/
	outfile = $1
    when /projection\s*=\s*(.*)/
	case $1
	when 'equalarea','a'
	    projection = Projection::Equalarea
	when 'equidistant','d'
	    projection = Projection::Equidistant
	when 'mercator','m'
	    projection = Projection::Mercator
	else
	    STDERR.print "Unknown projection #{projection_type}\n"
	    exit 1
	end
    else
	STDERR.print "Unknown line: #{line}\n"
	exit 1
    end
}

unless proj_n and proj_w
    STDERR.print "Projection parameters not specified\n"
    exit 1
end
$proj = projection.new (map_x, map_y, proj_n, proj_s, proj_w, proj_e)

if (not type_bm) or clabels
    arc=[]
    arcbb=[]
    open("pp/borders"){|fn|fn.each{|line|
	line =~ /(\d+) (\S+) (\S+) (\S+) (\S+) (.*)/
        arc.push $6.split(" ")
	arcbb.push Box.new($2.to_f, $3.to_f, $5.to_f, $4.to_f)
    }}
end

if type_bm
    system ("./project " + $proj.cmd + " >tmp-img.ppm")
    image = ImageList.new("tmp-img.ppm")

    open("pp/borders") {|fn| fn.each{|line|
	line =~ /(\d+) (\S+) (\S+) (\S+) (\S+) (.*)/
        next if $1.to_i < 4
        next unless $proj.bb_on_map? ($2.to_f, $3.to_f, $5.to_f, $4.to_f)
        d = Draw.new
	d.stroke_antialias true
        d.stroke_width 1.5
        d.stroke "gray60"
        d.fill "none"
        outline = $6.split(" ")
        outline2 = []
        outline.each {|xy|
	    x,y = xy.split(",")
	    x,y = $proj.project(x.to_f, y.to_f)
	    outline2.push x
	    outline2.push y
	}
        d.polyline(*outline2)
	d.draw image
    }}
else
    colors=[]
    open("pp/colors"){|fn|fn.each{|line|
	line =~ /(\d+) (\S+)/
        colors[$1.to_i]=$2
    }}

    image = Image.new(map_x, map_y) { self.background_color = "#000060" }

    nr = 0
    open("pp/pal"){|fn|fn.each{|line|
	line.chomp!
        ptx,pty = nil,nil
        holex,holey = nil,nil
        outline = []

        nr += 1
        next if nr == 1040

        catch (:dontrender) {
    	    p = line.split (" ").map{|z|z.to_i}
	    box = nil
	    p.each {|num|
		next if num == 0
	        b = arcbb[num.abs - 1]
		if box
		    box += b
		else
		    box = b
	        end
	    }
	    next unless $proj.bb_on_map?(box.l, box.r, box.u, box.d)
	    p.each{|num|
		if num == 0
		    throw :dontrender unless ptx
		    if holex
			outline.push holex
			outline.push holey
		    else
		        holex,holey = ptx,pty
		    end
		else
		    if num > 0
			c = arc[num-1]
		    else
	    		c = arc[-num-1].reverse
		    end
		    c.each {|ptd|
			x,y = ptd.split(",")
			ptx,pty = $proj.project(x.to_f, y.to_f)
		        outline.push ptx
			outline.push pty
		    }
		end
	    }
	    if holex
	        outline.push holex
		outline.push holey
	    end

	    d = Draw.new
#	    d.antialias true
	    d.stroke "none"
	    d.fill colors[nr]
	    d.polygon(*outline)
	    d.draw image
	}
    }}
end

if rivers
    open("pp/rivers") {|fn| fn.each{|line|
	line =~ /(\d+) (\S+) (\S+) (\S+) (\S+) (.*)/
        rank = $1.to_i
        break if rank >= 4

        next unless $proj.bb_on_map? ($2.to_f, $3.to_f, $5.to_f, $4.to_f)

        d = Draw.new
	d.stroke_antialias true
        if rank == 1
	    d.stroke_width 3
        elsif rank == 2
	    d.stroke_width 2
        else rank == 3
	    d.stroke_width 1.5
        end
        if type_bm
	    d.stroke "#0040FF"
	    d.opacity "50%"
        else
	    d.stroke "#000060"
        end
        d.fill "none"
        outline = $6.split(" ")
        outline2 = []
        outline.map {|xy|
	    x,y = xy.split(",")
	    x,y = $proj.project(x.to_f, y.to_f)
    	    outline2.push x
	    outline2.push y
        }
        d.polyline(*outline2)
        d.draw image
    }}
end

if cities
    labels=Labels.new(image)
    fn=open(cities)
    fn.each {|line|
	line =~ /^(.)\s+(.*?)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)/
        labels.add_city($1, $2, $4.to_f, $3.to_f)
    }
    labels.draw
end

if clabels
    lab=[]
    open("data/lab"){|fn|fn.each{|line|
	line =~ /(\d+) (.*)/
        lab[$1.to_i]=$2
    }}    

    country={}
    nr = 0
    open("pp/pal"){|fn|fn.each{|line|
	line.chomp!
        outline = []

        nr += 1
        next if nr == 1040
	next if lab[nr] == "?"

        
	p = line.split (" ").map{|z|z.to_i}
	box = nil
	p.each {|num|
	next if num == 0
	b = arcbb[num.abs - 1]
	    if box
	        box += b
	    else
	        box = b
	    end
	}
	next unless $proj.bb_on_map?(box.l, box.r, box.u, box.d)
	outline1=[]
	p.each{|num|
	    break if num == 0
	    if num > 0
	        c = arc[num-1]
	    else
	        c = arc[-num-1].reverse
	    end
	    c.each {|ptd|
	        x,y = ptd.split(",")
	        ptx,pty = $proj.project(x.to_f, y.to_f)
	        outline1.push [ptx,pty]
	    }
	}
	outline2 = []
	outline1.each_pair2 {|x,y,lpx,lpy|
	    outline2.push [0, lpy + (y-lpy) * lpx / (lpx-x)] if (x < 0) != (lpx < 0)
	    outline2.push [x,y] if x >= 0
	}
	outline3 = []
	outline2.each_pair2 {|x,y,lpx,lpy|
	    outline3.push [lpx + (x-lpx) * lpy / (lpy-y), 0] if (y < 0) != (lpy < 0)
	    outline3.push [x,y] if y >= 0
	}
	    outline4 = []
	outline3.each_pair2 {|x,y,lpx,lpy|
	    outline4.push [map_x, lpy + (y-lpy) * (lpx-map_x) / (lpx-x)] if (x > map_x) != (lpx > map_x)
	    outline4.push [x,y] if x <= map_x
	}
	outline5 = []
	outline4.each_pair2 {|x,y,lpx,lpy|
	    outline5.push [lpx + (x-lpx) * (lpy-map_y) / (lpy-y), map_y] if (y > map_y) != (lpy > map_y)
	    outline5.push [x,y] if y <= map_y
	}	    
	area = 0.0
	cx,cy = 0.0, 0.0
	outline5.each_pair2 {|x,y,lpx,lpy|
	    sf = lpx * y - x * lpy
	    area += sf
	    cx += (x + lpx) * sf
	    cy += (y + lpy) * sf
	}
	area /= 2
	next if area == 0.0
	    
	cx /= 6 * area
	cy /= 6 * area

	country_name = lab[nr]
	country[country_name] ||= [0.0, 0.0, 0.0]
	x0, y0, a0 = country[country_name]
	country[country_name] = [x0 + cx * area, y0 + cy * area, a0 + area]
    }}

    oversea={}
    open("data/oversea"){|fn|fn.each{|line|
        line =~ /(.*?)\t+(.*)/
        oversea[$1]=$2
    }}

    locale_cnames={}
    if locale
	open("data/clocale.#{locale}"){|fn|fn.each{|line|
	    line =~ /(.*?)\t+(.*)/
	    locale_cnames[$1]=$2
	}}
    end    

    country.each{|country_name,(x,y,a)|
        x /= a
        y /= a

	lcname        = locale_cnames[country_name] || country_name
	mainland_name = oversea[country_name]
	lmname        = locale_cnames[mainland_name] || mainland_name if mainland_name

        text = Draw.new
        text.text_antialias = true
	if type_bm
	    text.fill = "#B350B350FFFF8000"
	else
	    text.fill = '#FF8080'
	end
        text.font_weight = BoldWeight
	if a > 200000
    	    text.pointsize = 72
	elsif a > 100000
    	    text.pointsize = 48
	elsif a > 30000
	    text.pointsize = 36
	elsif a > 10000
	    text.pointsize = 24
	elsif a > 3000
	    text.pointsize = 18
	else
    	    text.pointsize = 12
	end
	if lmname
	    lmname = "(#{lmname})"
	    metrics1 = text.get_type_metrics(image, lcname)
	    metrics2 = text.get_type_metrics(image, lmname)
	    text.annotate(image, 0, 0, x - metrics1.width/2, y, lcname)
	    text.annotate(image, 0, 0, x - metrics2.width/2, y - metrics1.descent + metrics2.ascent, lmname)
	else
	    metrics = text.get_type_metrics(image, lcname)
	    d,u = y + metrics.ascent/2, y - metrics.ascent/2
	    l,r = x - metrics.width/2,  x + metrics.width/2
	    if l < 0
		l = 0
		r = l + metrics.width
	    end
	    if r > map_x
		r = map_x
		l = r - metrics.width
	    end
	    if u < 0
		u = 0
		d = l + metrics.ascent
	    end
	    if d > map_y
		d = map_y
		u = d - metrics.ascent
	    end
	    text.annotate(image, 0, 0, l, d, lcname)
	end
    }
end

image.write(outfile)
